// @ myth - common - map.C
//
// Myth is a multiuser screen oriented adventure game.
// Copyright (C) 1989 Per Abrahamsen
// 
// This file is part of Myth.
// 
// Myth is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 1, or (at your option) any later
// version.
// 
// Myth is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Myth; see the file COPYING.  If not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
// My email address is: amanda@iesd.auc.dk

// @@ include

#include "common/map.h"
#include "common/error.h"

// @@ class map_line

// @@@ void map_line::send(int start, int line, int stop, BINARY* binary)

void map_line::send(int start, int line, int stop, BINARY* binary)
{
    binary->PutCoor(coor(start, line));
    for(int i = start; i <= stop; i++)
    {
	if(data[i - first] == 0)
	{
	    bug.error("map_line::send", "zero is reserved");
	    binary->PutChar(127);
	}
	else
	    binary->PutChar(data[i - first]);
    }
    binary->PutChar(0);
}    

// @@@ void map_line::assign(int x,

void map_line::assign(int x, int length, const char* buffer, char EmptyMap)
{
    // Create new entry
    if(data == NULL)
    {
	first = x - ExtraMap;
        last = x + length + ExtraMap;
	data = new char[last - first + 1];

	for(int i = first; i <= last; i++)
	{
	    if(i >= x && i < x + length)
		data[i - first] = buffer[i - x];
	    else
		data[i - first] = EmptyMap;
	}
    }
    // Fit into current entry
    else if(x >= first && x + length - 1 <= last)
    {
	for(int i = 0; i < length; i++)
	    data[i + (x - first)] = buffer[i];
    }
    // Resize entry
    else
    {
	int from = (x < first) ? x - ExtraMap : first;
	int to = (x + length > last) ? x + length + ExtraMap : last;

	char* old = data;
	data = new char[to - from + 1];
	for(int i = from; i <= to; i++)
	{
	    if(i >= x && i < x + length)
		data[i - from] = buffer[i - x];
	    else if(i < first || i > last)
		data[i - from] = EmptyMap;
	    else
		data[i - from] = old[i - first];
	}
	delete old;
	first = from;
	last = to;
    }
}

// @@@ const map_line& map_line::operator=(map_line line)

const map_line&  map_line::operator=(const map_line& line)
{
    delete data;
    first = line.first;
    last = line.last;
    data = strncpy(new char[last - first + 1], line.data, last - first + 1);
    return line;
}

// @@@ void map_line::save(BINARY* binary, int line, char EmptyMap)

void map_line::save(BINARY* binary, int line, char EmptyMap)
{
    // The cost in bytes to split up a line in two fragments.
    static const int SizeOfLineHeader = 5;
    int start = first;		// Start of area to send to the buffer
    int empty = 0;		// Empty chars in the tail

    for(int column = first; column <= last; column++)
    {
	if(data[column - first] == EmptyMap)
	{
	    if(start == column)
		start++;
	    else if(++empty > SizeOfLineHeader)
	    {
		send(start, line, column - empty, binary);
		empty = 0;
		start = column + 1;
	    }
	}
	else
	{
	    if(column - start > MessageSize - SizeOfLineHeader - 5)
	    {
		send(start, line, column, binary);
		start = column + 1;
	    }
	    empty = 0;
	}
    }
    if(start <= last)
	send(start, line, last - empty, binary);
}

// @@@ map_line::map_line()

map_line::map_line()
{
    first = 0;
    last = -1;
    data = NULL;
}

// @@@ map_line::~map_line()

map_line::~map_line()
{
    delete data;
}

// @@ class MAP

// @@@ void MAP::resize(int from, int to)

void MAP::resize(int from, int to)
{
    if(data == NULL)
	data = new map_line[to - from + 1];
    else
    {
	map_line* old = data;
	data = new map_line[to - from + 1];
	for(int i = first; i <= last; i++)
	    data[i - from] = old[i - first];
	delete old;
    }
    first = from;
    last = to;
}

// @@@ void MAP::assign(coor pos, int length, const char* buffer)

void MAP::assign(coor pos, int length, const char* buffer)
{
    if(data == NULL)
	resize(pos.y - ExtraMap, pos.y + ExtraMap);
    else if(pos.y < first)
	resize(pos.y - ExtraMap, last);
    else if(pos.y > last)
	resize(first, pos.y + ExtraMap);
    
    data[pos.y - first].assign(pos.x, length, buffer, EmptyMap);
}

// @@@ void MAP::clear()

void MAP::clear()
{
    if(data != NULL)
    {
	delete data;
	data = NULL;
    }
    first = 0;
    last = -1;
}

// @@@ void MAP::save(BINARY* binary)

void MAP::save(BINARY* binary)
{
    for(int line = first; line <= last; line++)
	data[line - first].save(binary, line, EmptyMap);

    binary->PutCoor(NoWhere);
}

// @@@ void MAP::load(BINARY* binary)

void MAP::load(BINARY* binary)
{
    coor pos;

    // Bug in G++ prevent: while((pos = binary->GetCoor()) != NoWhere)
    pos = binary->GetCoor();
    
    while(pos != NoWhere)
    {
	print(pos, binary->GetString());
	pos = binary->GetCoor();
    }
}

// @@@ void MAP::SaveDefault(BINARY* binary)

void MAP::SaveDefault(BINARY* binary)
{
    binary->PutChar(EmptyMap);
}

// @@@ void MAP::LoadDefault(BINARY* binary)

void MAP::LoadDefault(BINARY* binary)
{
    EmptyMap = binary->GetChar();
}

// @@@ MAP::MAP()

MAP::MAP()
{
    EmptyMap = ' ';
    first = 0;
    last = -1;
    data = NULL;
}

// @@@ MAP::~MAP()

MAP::~MAP()
{
    delete data;
}

// @@ emacs
//
// Local Variables:
// mode: c++
// outline-regexp: "// @+ "
// eval: (outline-minor-mode 1)
// End:
