// @ myth - server - level.C
//
// Myth is a multiuser screen oriented adventure game.
// Copyright (C) 1989 Per Abrahamsen
// 
// This file is part of Myth.
// 
// Myth is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 1, or (at your option) any later
// version.
// 
// Myth is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Myth; see the file COPYING.  If not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
// My email address is: amanda@iesd.auc.dk

// @@ include

#include "server/level.h"
#include "common/error.h"
#include "common/binary.h"
#include "common/parse.h"
#include "server/player.h"
#include "server/rnd.h"
#include "common/format.h"

// @@ frequence_link

// @@@ void frequence_link::save(BINARY* binary)

void frequence_link::save(BINARY* binary)
{
    binary->PutShort(object);
    binary->PutShort(chance);
}

// @@@ frequence_link::frequence_link(BINARY* binary)

frequence_link::frequence_link(BINARY* binary)
{
    object = binary->GetShort();
    chance = binary->GetShort();
}

// @@@ frequence_link::frequence_link(object_type type, int when)

frequence_link::frequence_link(object_type type, int when)
{
    object = type;
    chance = when;
}

// @@ FREQUENCE

// @@@ void FREQUENCE::erase()

void FREQUENCE::erase()
{
    delete FREQUENCE::current();
    LIST::remove();
}

// @@@ void FREQUENCE::clear()

void FREQUENCE::clear()
{
    FREQUENCE::reset();

    while(FREQUENCE::check())
	FREQUENCE::erase();
}

// @@@ void FREQUENCE::add(object_type object, int chance)

void FREQUENCE::add(object_type object, int chance)
{
    LIST::add(new frequence_link(object, chance));
    all_monsters += chance;
}

// @@@ MONSTER* FREQUENCE::CreateMonster(int value)

MONSTER* FREQUENCE::CreateMonster(int value)
{
    int where = rnd(1, all_monsters);
    
    reset();
    while(where > chance())
    {
	where -= chance();
	next();
    }
    object_type what = type();
    
    MONSTER* monster = (MONSTER*) CreateObject(what);
    
    switch(what)
    {
    case MONSTER_RAT:
	monster->SetExperience(1 + value / 2);
	break;
    case MONSTER_BAT:
	monster->SetExperience(5 + value / 4);
	break;
    default:
	/* Do Nothing */
	break;
    }
    return monster;
}

// @@@ bool FREQUENCE::MonsterChance()

bool FREQUENCE::MonsterChance()
{ return rnd.chance(1, chance_monster); }

// @@@ bool FREQUENCE::ObjectChance()

bool FREQUENCE::ObjectChance()
{ return rnd.chance(1, chance_object); }

// @@@ bool FREQUENCE::NewMonster()

bool FREQUENCE::NewMonster()
{ return rnd.chance(1, chance_create); }

// @@@ char FREQUENCE::locate()

char FREQUENCE::locate()
{ return where.string()[rnd(0, where.length() - 1)]; }

// @@@ void FREQUENCE::save(BINARY* binary)

void FREQUENCE::save(BINARY* binary)
{
    binary->PutShort(chance_monster);
    binary->PutShort(chance_object);
    binary->PutShort(chance_create);
    binary->PutString(where());
    
    reset();

    while(check())
    {
	binary->PutChar(CONTINUE);
	current()->save(binary);
	next();
    }
    binary->PutChar(DONE);
}

// @@@ void FREQUENCE::load(BINARY* binary)

void FREQUENCE::load(BINARY* binary)
{
    chance_monster = binary->GetShort();
    chance_object = binary->GetShort();
    chance_create = binary->GetShort();
    where = binary->GetString();

    while(binary->GetChar() == CONTINUE)
    {
	frequence_link* link = new frequence_link(binary);

	LIST::add(link);
	all_monsters += link->chance;
    }
}

// @@@ void FREQUENCE::load(PARSE* parse)

void FREQUENCE::load(PARSE* parse)
{
    chance_monster = parse->GetLong();
    parse->skip(INPUT_SPACE);
    chance_object = parse->GetLong();
    parse->skip(INPUT_SPACE);
    chance_create = parse->GetLong();
    parse->skip(INPUT_SPACE);
    where = STRING(parse->GetString());
    parse->skip(INPUT_COMMENT);
    
    object_type type = ObjectToType(parse->GetWord());
    
    while(type != OBJECT_END && parse->good())
    {
	parse->skip(INPUT_SPACE);
	int chance = parse->GetLong();
	add(type, chance);
	parse->skip(INPUT_COMMENT);
	type = ObjectToType(parse->GetWord());
    }
}

// @@@ FREQUENCE::FREQUENCE()

FREQUENCE::FREQUENCE()
{
    chance_create = chance_object = chance_monster = 0;
    all_monsters = 0;
}

// @@ BUCKET

// @@@ BUNCH* BUCKET::what(coor pos)

BUNCH* BUCKET::what(coor pos)
{
    BunchList.reset();
    PositionList.reset();

    while(BunchList.current() != NULL)
    {
	if(*((coor*) PositionList.current()) == pos)
	    return (BUNCH*) BunchList.current();

	BunchList.next();
	PositionList.next();
    }
    return NULL;
}

// @@@ BUNCH* BUCKET::create(coor pos)

BUNCH* BUCKET::create(coor pos)
{
    BUNCH* bunch = what(pos);
    
    if(bunch == NULL)		// If no bunch exist, create one.
    {
	bunch = new BUNCH;
	BunchList.add(bunch);
	PositionList.add(new coor(pos));
    }
    return bunch;
}

// @@@ void BUCKET::update()

void BUCKET::update()
{
    BunchList.reset();
    PositionList.reset();

    while(BunchList.current() != NULL)
    {
	BUNCH* bunch = (BUNCH*) BunchList.current();
	coor* pos = (coor*) PositionList.current();

	if(bunch->empty())
	{
	    BunchList.remove();
	    delete bunch;
	    PositionList.remove();
	    delete pos;
	}
	else
	{
	    BunchList.next();
	    PositionList.next();
	}
    }
}

// @@@ void BUCKET::save(BINARY* binary)

void BUCKET::save(BINARY* binary)
{
    BunchList.reset();
    PositionList.reset();

    while(BunchList.current() != NULL)
    {
	BUNCH* bunch = (BUNCH*) BunchList.current();
	coor* pos = (coor*) PositionList.current();

	binary->PutCoor(*pos);
	bunch->save(binary);

	BunchList.next();
	PositionList.next();
    }
}

// @@@ void BUCKET::free()

void BUCKET::free()
{
    BunchList.reset();
    PositionList.reset();

    while(BunchList.current() != NULL)
    {
	BUNCH* bunch = (BUNCH*) BunchList.current();
	coor* pos = (coor*) PositionList.current();

	bunch->free();
	BunchList.remove();
	delete bunch;
	PositionList.remove();
	delete pos;
    }
}    

// @@ EVENT

// @@@ void EVENT::erase()

void EVENT::erase()
{
    delete current();
    list.remove();
}

// @@@ void EVENT::clear()

void EVENT::clear()
{
    reset();
    
    while(current() != NULL)
	erase();
}

// @@@ void EVENT::add(coor pos, long when)

void EVENT::add(coor pos, long when)
{
    reset();
    while(current() != NULL)
    {
	if(current()->when < when)
	    next();
	else
	{
	    insert(pos, when);
	    return;
	}
    }
    insert(pos, when);
}

// @@@ void EVENT::update(LEVEL* level)

void EVENT::update(LEVEL* level)
{
    while(current() != NULL)
    {
	if(current()->when > watch())
	    return;
	else
	{
	    coor pos = current()->pos;
	    BUNCH* bunch = level->create(pos);
	    bunch->update(level, pos, NULL);
	    erase();
	}
    }
}

// @@@ void EVENT::save(BINARY* binary)

void EVENT::save(BINARY* binary)
{
    reset();
    while(current() != NULL)
    {
	binary->PutLong(current()->when);
	binary->PutCoor(current()->pos);
	next();
    }
    binary->PutLong(0L);
}

// @@@ void EVENT::load(BINARY* binary)

void EVENT::load(BINARY* binary)
{
    long when = binary->GetLong();

    while(when != 0L)
    {
	add(binary->GetCoor(), when);
	when = binary->GetLong();
    }
}

// @@ LEVEL

// @@@ void LEVEL::AddMonster(MONSTER* monster)

void LEVEL::AddMonster(MONSTER* monster)
{
    zoo.add(monster);
}

// @@@ void LEVEL::AddPlayer(PLAYER* player)

void LEVEL::AddPlayer(PLAYER* player)
{ party.add(player); }

// @@@ err LEVEL::restore(id_type id, PLAYER* player)

err LEVEL::restore(id_type id, PLAYER* player)
{
    // Make sure this player object is closed and freeded after use
    AddMonster(player);
    AddPlayer(player);
    
    // Find the sleeping player
    sleepers.reset();
    while(sleepers.remain() > 0)
    {
	PLAYER* current = sleepers.current();
	
	if(current->identity() == id)
	{
	    // Remove it from the monster queue
	    zoo.reset();
	    while(zoo.remain() > 0)
	    {
		if(zoo.current() == current)
		    zoo.remove();
		else
		    zoo.next();
	    }
	    
	    // Check for valid position
	    if(current->position() == NoWhere)
	    {
		sleepers.erase();
		player->message("Sorry, i don't see you around here!");
		return ERROR;
	    }

	    // Remove from bunch
	    SPECIFY_OBJECT specify(current);
	    OBJECT_LIST list;
	    BUNCH* bunch = create(current->position());
	    bunch->remove(&specify, &list, NULL, -1L);

	    // He should not be sleeping here any more!
	    sleepers.remove();

	    switch(list.length())
	    {
	    case 1:
		if(current->dead())
		{
		    player->message("Sorry, someone has killed you!");
		    bunch->add(current);
		}
		else
		{
		    current->sleep(this, -60L * 10L);
		    current->connect(player->GetChannel());
		    current->enter(this);
		    return SUCCES;
		}
		break;
	    case 0:
		bug.error("LEVEL::restore", "Zero objects");
		player->message("Sorry, your body has disapeared.");
		break;
	    default:
		bug.error("LEVEL::restore", "Multiple objects");
		player->message("You seem to be everywhere, but...");
	    }
	    return ERROR;
	}
	sleepers.next();
    }
    player->message("Your soul is unable to find your physical body.");
    return ERROR;
}

// @@@ void LEVEL::GetMonster(MONSTER_LIST* list, coor pos)

void LEVEL::GetMonster(MONSTER_LIST* list, coor pos)
{
    zoo.reset();
    while(zoo.remain() > 0)
    {
	if(zoo.current()->position() == pos)
	    list->add(zoo.current());
	
	zoo.next();
    }
}    

// @@@ PLAYER* LEVEL::FindPlayer(MONSTER* monster)

PLAYER* LEVEL::FindPlayer(MONSTER* monster)
{
    PLAYER* BestPlayer = NULL;
    int BestSquare = sqr(5 + experience()); // Depend on monster->smell() ?
    const coor center = monster->position();
    PLAYER* player;

    party.reset();
    while((player = party.current()) != NULL)
    {
	coor pos = player->position();
	int square = sqr(center.x - pos.x) + sqr(center.y - pos.y);

	if(square < BestSquare)
	{
	    BestSquare = square;
	    BestPlayer = player;
	}
	party.next();
    }
    return BestPlayer;
}

// @@@ MONSTER* LEVEL::FindEnemy(MONSTER* monster)

MONSTER* LEVEL::FindEnemy(MONSTER* monster)
{
    MONSTER* BestEnemy = NULL;
    int BestSquare = sqr(25 + experience()); // Depend on monster->smell() ?
    const coor center = monster->position();
    MONSTER* enemy;

    zoo.reset();
    while((enemy = zoo.current()) != NULL)
    {
	if(monster->IsEnemy(enemy))
	{
	    coor pos = enemy->position();
	    int square = sqr(center.x - pos.x) + sqr(center.y - pos.y);
	    
	    if(square < BestSquare)
	    {
		BestSquare = square;
		BestEnemy = enemy;
	    }
	}
	zoo.next();
    }
    return BestEnemy;
}

// @@@ BUNCH* LEVEL::bunch(coor pos)

BUNCH* LEVEL::bunch(coor pos)
{
    return bucket[abs(pos.x) % BucketSize][abs(pos.y) % BucketSize].what(pos);
}

// @@@ BUNCH* LEVEL::create(coor pos)

BUNCH* LEVEL::create(coor pos)
{
    return
	bucket[abs(pos.x) % BucketSize][abs(pos.y) % BucketSize].create(pos);
}

// @@@ void LEVEL::update(coor pos)

void LEVEL::update(coor pos)
{
    bucket[abs(pos.x) % BucketSize][abs(pos.y) % BucketSize].update();
}

// @@@ void LEVEL::AddEvent(coor pos, long when)

void LEVEL::AddEvent(coor pos, long when)
{ event.add(pos, when); }

// @@@ bool LEVEL::lit(coor pos)

bool LEVEL::lit(coor pos)
{ return (map.what(pos) & MAP_MASK_LIT) != 0; }

// @@@ void LEVEL::light(coor pos)

void LEVEL::light(coor pos)
{ map.plot(pos, map.what(pos) | MAP_MASK_LIT); }

// @@@ void LEVEL::dark(coor pos)

void LEVEL::dark(coor pos)
{ map.plot(pos, map.what(pos) & ~MAP_MASK_LIT); }

// @@@ rock_type LEVEL::rock(coor pos)

rock_type LEVEL::rock(coor pos)
{ return (map.what(pos) & MAP_MASK_ROCK) >> MAP_OFFSET_ROCK; }

// @@@ void LEVEL::dig(coor pos, rock_type rock)

void LEVEL::dig(coor pos, rock_type rock)
{
    map.plot(pos,
	     ((map.what(pos) & ~MAP_MASK_ROCK) | (rock << MAP_OFFSET_ROCK)));
}

// @@@ char LEVEL::symbol(coor pos, bool SeeInvisible)

char LEVEL::symbol(coor pos, bool SeeInvisible)
{
    rock_type rock = LEVEL::rock(pos);
    
    if(!IsRock(rock))
    {
	BUNCH* bunch = LEVEL::bunch(pos);
	
	if(bunch == NULL)
	    /* Do Nothing... */;
	else if(bunch->empty())
	    update(pos);
	else
	{
	    char c = bunch->look(SeeInvisible);

	    if(c != SYMBOL_ERROR)
		return c;
	}
    }
    return RockSymbol[rock];
}

// @@@ bool LEVEL::IsLit(coor pos)

bool LEVEL::IsLit(coor pos)
{
    if(lit(pos))
	return TRUE;
    else
    {
	BUNCH* heap = bunch(pos);

	return heap != NULL && heap->light();
    }
}

// @@@ void LEVEL::WatchMonster(MONSTER* monster,

void LEVEL::WatchMonster(MONSTER* monster, const char* string,
			 OBJECT* object = NULL)
{
    const coor pos = monster->position();
    STRING message;
    if(object == NULL)
	message = format(string, print(monster, PRINT_MONSTER | PRINT_UPPER));
    else
    {
	message = format(string, print(monster, PRINT_MONSTER | PRINT_UPPER),
			 print(object, PRINT_COUNT));
    }
    PLAYER* player;

    party.reset();
    while((player = party.current()) != NULL)
    {
	if(player->position() == pos || player->CanSee(pos))
	{
	    if(player != monster)
		player->message(message);
	}
	party.next();
    }
}

// @@@ void LEVEL::WatchPosition(coor pos, STRING string)

void LEVEL::WatchPosition(coor pos, STRING string)
{
    PLAYER* player;

    party.reset();
    while((player = party.current()) != NULL)
    {
	if(player->position() == pos || player->CanSee(pos))
	    player->message(string);

	party.next();
    }
}

// @@@ void LEVEL::WatchPosition(coor pos, const char* string)

void LEVEL::WatchPosition(coor pos, const char* string)
{ WatchPosition(pos, STRING(string)); }

// @@@ void LEVEL::WatchEffect(coor pos, char c)

void LEVEL::WatchEffect(coor pos, char c)
{
    PLAYER* player;

    party.reset();
    while((player = party.current()) != NULL)
    {
	if(player->position() == pos || player->CanSee(pos))
	    player->SpecialEffect(pos, c);

	party.next();
    }
}    

// @@@ void LEVEL::hiss(MONSTER* hissing)

void LEVEL::hiss(MONSTER* hissing)
{
    hissing->message("You hiss!");
    WatchMonster(hissing, "%s is about to hiss...");

    const coor center = hissing->position();
    MONSTER* monster;

    zoo.reset();
    while((monster = zoo.current()) != NULL)
    {
	coor pos = monster->position();
	int chance = sqr(hissing->experience()) + 1 -
	            (sqr(center.x - pos.x) + sqr(center.y - pos.y));

	if(chance > 0)
	{
	    if(hissing != monster)
		monster->message("You hear a hissing sound.");
	    
	    if(!monster->CheckAbility(ABILITY_STONE_PROTECTION) &&
	       !monster->CheckAbility(ABILITY_PHASING) &&
	       !monster->CheckAbility(ABILITY_STONED) &&
	       rnd.chance(1, chance + 2))
		monster->AddAbility(new COUNTER_STONED(5L));
	}
	zoo.next();
    }
}

// @@@ void LEVEL::talk(MONSTER* talker, const char* what)

void LEVEL::talk(MONSTER* talker, const char* what)
{
    const coor where = talker->position();
    PLAYER* player;

    party.reset();
    while((player = party.current()) != NULL)
    {
	coor pos = player->position();

	if(player == talker)
	    player->message(format("You say:%s", what));
	else if(player->CanSee(where))
	    player->message(format("%s says:%s",
				   ::print(talker, PRINT_MONSTER), what));
	else if(sqr(talker->strength() + player->dexterity()) >
		sqr(where.x - pos.x) + sqr(where.y - pos.y))
	    player->message(format("You hear:%s", what));

	party.next();
    }
}

// @@@ coor LEVEL::locate(int l)

coor LEVEL::locate(int l)
{
    if(l >= 0 && l < MaxChar)
	return location[l];
    else
	return NoWhere;
}

// @@@ LEVEL::LEVEL(level_type type)

LEVEL::LEVEL(level_type type)
{
    // Initialise
    int i;
    
    LastVisit = 0;
    CountDown = MaxLevelLife;
    (void) strncpy(RockSymbol, ".,######", ROCK_END);
    MyDepth = type;
    MyExperience = 5;
    SaveChance = 750;
    
    for(i = 0; i <BucketSize; i++)
	bucket[i] = new BUCKET[BucketSize];
    for(i = 0; i < MaxChar; i++)
	location[i] = NoWhere;

    map.SelectDefault(MAP_MASK_LIT | (ROCK << MAP_OFFSET_ROCK));

    // Load
    BINARY binary(BUFFER_READ, global.CaveBinaryFile(type).string());
    
    if(binary.GoodInput())
	load(&binary);
    else
    {
	PARSE parse(global.CaveTextFile(type).string());

	if(parse.good())
	    LoadText(&parse);
	else
	    bug.error("LEVEL::LEVEL",
		      format("Can't open %s",
			     global.CaveTextFile(type).string()));
    }
}

// @@@ LEVEL::~LEVEL()

LEVEL::~LEVEL()
{
    // Free all objects
    for(region reg(Origo, coor(BucketSize, BucketSize));
	reg.check(); reg.NextX())
    {
	bucket[reg.x][reg.y].free();
    }
}

// @@ emacs
//
// Local Variables:
// mode: c++
// outline-regexp: "// @+ "
// eval: (outline-minor-mode 1)
// End:
