// @ myth - server - mythd.C
//
// Myth is a multiuser screen oriented adventure game.
// Copyright (C) 1989 Per Abrahamsen
// 
// This file is part of Myth.
// 
// Myth is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 1, or (at your option) any later
// version.
// 
// Myth is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Myth; see the file COPYING.  If not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
// My email address is: amanda@iesd.auc.dk

// @@ include

#include "common/command.h"
#include "common/wait.h"
#include "server/watch.h"
#include "common/list.h"
#include "server/player.h"
#include "server/question.h"
#include "server/hiscore.h"
#include "server/sleep.h"
#include "common/error.h"
#include "server/global.h"
#include "common/format.h"
#include <signal.h>

// @@ forward

static void CatchPhone();

// @@ const

// How many fight point is given at each tick
static const int TickFight = 100;
// How many move point is given at each tick
static const int TickMove = 100;

// @@ LEVEL_LIST

// @@@ class LEVEL_LIST

class LEVEL_LIST
{
    LIST list;
    LEVEL** table;	// Table of levels
public:
    // These are used to loop through all active levels
    void inline reset()
    { list.reset(); }
    inline LEVEL* current()
    { return (LEVEL*) list.current(); }
    inline void next()
    { list.next(); }
    // Remove the current level
    void erase();
    // Return a pointer to a level, creating it if nonexistent.
    LEVEL* find(level_type type);
    // Check if any levels is active
    inline bool empty()
    { return list.length() < 1; }
    // Remove all active levels
    void clear();
    // Initialise the table
    LEVEL_LIST();
};

// @@@ void LEVEL_LIST::erase()

void LEVEL_LIST::erase() 
{
    table[current()->depth()] = NULL;
    delete current();
    list.remove();
}

// @@@ LEVEL* LEVEL_LIST::find(level_type type)

LEVEL* LEVEL_LIST::find(level_type type)
{
    if(table[type] == NULL)
    {
	table[type] = new LEVEL(type);
	list.add(table[type]);
    }
    return table[type];
}

// @@@ void LEVEL_LIST::clear()

void LEVEL_LIST::clear()
{
    reset();
	
    while(!empty())
	erase();
}    

// @@@ LEVEL_LIST::LEVEL_LIST()

LEVEL_LIST::LEVEL_LIST()
{
    table = new LEVEL*[global.DungeonSize()];
    
    for(level_type i = 0; i < global.DungeonSize(); i++)
	table[i] = NULL;
}

// @@ void ParseNewPlayer(PLAYER_LIST* news, MONSTER_LIST* export)

void ParseNewPlayer(PLAYER_LIST* news, MONSTER_LIST* export)
{
    news->reset();

    while(news->remain() > 0)
    {
	PLAYER* player = news->current();
	
	if(!player->good())
	{
	    bug.warning("::ParseNewPlayer", "bad I/O for player");
	    news->erase();
	}
	else
	{
	    player->parse();
	    
	    switch(player->answer(NULL))
	    {
	    case Q_VALUE_PARSING:
		news->next();
		break;
	    case Q_VALUE_DONE:
		player->retry();
		
		if(player->busy())
		{
		    player->refresh();
		    news->next();
		}
		else
		{
		    export->add(player);
		    news->remove();
		}
		break;
	    case Q_VALUE_AGAIN:
		// No "AutoRun" before she start to play
	    case Q_VALUE_ERROR:
	    default:
		news->erase();
		bug.error("::ParseNewPlayer", "answer is garbage");
	    }
	}
    }
}    

// @@ DUNGEON

// @@@ class DUNGEON
//
// This class contain most of the game.

class DUNGEON
{
    // File to save the state in
    STRING name;
    STRING backup;
    // Keep some levels in memory
    LEVEL_LIST list;		// List of levels
    // Remember which players is sleeping where
    SLEEP sleep;
    // Keep track of who's who in this game
    HI_SCORE HiScore;
    // Update IsReady for each player.
    cont check(WAIT* wait, bool sleeping, bool AutoRun);
    // Move monsters between levels
public:
    void MoveMonsters(MONSTER_LIST* export, WAIT* wait);
private:
    // Remember the games state
    void save();
public:
    void play(GATE* gate);
    DUNGEON();
    ~DUNGEON();
};

// @@@ cont DUNGEON::check(WAIT* wait, bool sleeping, bool AutoRun)

cont DUNGEON::check(WAIT* wait, bool sleeping, bool AutoRun)
{
    if(sleeping)
    {
	// Nobody is playing
	save();
	// The game can safely be killed now!	    
	wait->check(global.TimeOut());
	watch.update();
	return DONE;
    }
    else if(AutoRun)
    {
	// Just check if some other have data ready
	wait->check(0);
    }
    else
    {
	// No active players, wait for one
	wait->check(TickTime);
	watch.update();
    }
    return CONTINUE;
}    

// @@@ void DUNGEON::MoveMonsters(MONSTER_LIST* export, WAIT* wait)

void DUNGEON::MoveMonsters(MONSTER_LIST* export, WAIT* wait)
{
    export->reset();

    while(export->remain() > 0)
    {
	MONSTER* monster = export->current();
	
	if(monster->dead() || monster->escape())
	    // Give the monster credit for what it has done
	    HiScore.add(monster->ScoreEntry());
	
	if(monster->done() || monster->escape())
	    // This object is not used any more
	    export->erase();
	else if(monster->dead())
	    // Dead and gone, but leave the body around
	    export->remove();
	else if(global.emergency())
	{
	    monster->message("The stairway collapse around you!");
	    monster->kill("was trapped in a stairway during an earthquake",
			  NULL);
	    // No export->erase, take care of it next time...
	}
	else
	{
	    // Move to a new level
	    LEVEL* level = list.find(monster->depth());
	    monster->enter(level);
	    (void) level->play(export, wait, &sleep, &HiScore, FORCE);
	    export->remove();
	}
    }
}

// @@@ void DUNGEON::save()

void DUNGEON::save()
{
    (void) unlink(backup());
    (void) link(name(), backup());
    (void) unlink(name());

    BINARY binary(BUFFER_WRITE, name());

    ::identity.save(&binary);
    sleep.save(&binary);
    HiScore.save(&binary);

    if(!binary.GoodOutput())
    {
	bug.warning("DUNGEON::~DUNGEON", "Can't save state");
	(void) unlink(name());
	(void) link(backup(), name());
    }    
}

// @@@ void DUNGEON::play(GATE* gate)
//
// This is the main game loop.

// A strange bug in g++ 1.36.1 on a sun3
static DUNGEON* ugly_ugly;

void DUNGEON::play(GATE* gate)
{
    // Examine what players have input ready
    WAIT wait;
    wait.add(gate);		// And new players from the gate

    // TRUE, if any player is automaticly moving
    bool AutoRunFlag = FALSE;

    // List of new players, which has not been placed in the game
    PLAYER_LIST news;

    while(ugly_ugly->check(&wait, bool(ugly_ugly->list.empty() && news.length() < 1),
		AutoRunFlag) == CONTINUE ||
	  gate->IsReady)
	  
    {
	// In case we got a SIGHUP
	bool TheEndOfTheWorld = global.emergency();

	// Clean up temporaries for formating
	FormatCleanUp();

	AutoRunFlag = FALSE;

	// Monsters on their way out
	MONSTER_LIST export;

	// Add new players from the gate
	if(gate->IsReady)
	    news.add(new PLAYER(gate, &wait, &ugly_ugly->sleep));

	// Let new players try select a game
	ParseNewPlayer(&news, &export);

	// Update each level
	ugly_ugly->list.reset();
	LEVEL* level;
	while((level = ugly_ugly->list.current()) != NULL)
	{
	    // Clean up temporaries for formating
	    FormatCleanUp();
	    
	    AutoRunFlag |= level->play(&export, &wait, &ugly_ugly->sleep,
						&ugly_ugly->HiScore, CHECK);

	    if(level->timeout() || TheEndOfTheWorld)
	    {
		MONSTER_LIST zoo;

		level->CheckOut(&zoo, &wait, &ugly_ugly->sleep, &ugly_ugly->HiScore);
		ugly_ugly->MoveMonsters(&zoo, &wait);
		ugly_ugly->list.erase();

		if(!TheEndOfTheWorld)
		    ugly_ugly->save();
	    }
	    else
		ugly_ugly->list.next();
	}

	// Place monsters which has changed level
	ugly_ugly->MoveMonsters(&export, &wait);

	if(TheEndOfTheWorld)
	    break;
    }
}

// @@@ DUNGEON::DUNGEON()

DUNGEON::DUNGEON()
{
    name = format("%s/%s/%s", HomeDirectory, LibraryDirectory, "state");
    backup = format("%s.bak", name());
    
    BINARY binary(BUFFER_READ, name());
	
    if(binary.GoodInput())
    {
	::identity.load(&binary);
	sleep.load(&binary);
	HiScore.load(&binary);
    }
}

// @@@ DUNGEON::~DUNGEON()

DUNGEON::~DUNGEON()
{
    // Save the state
    save();
    // Save all levels
    list.clear();
}

// @@ int main(int argc, char** argv)

int main(int argc, char** argv)
{
    bug.open(format("%s/%s/ERR.%s", HomeDirectory, ErrorDirectory, "daemon"));
    global.load(argc, argv);
    GATE gate;	// Accept connections from clients
    (void) signal(SIGPIPE, SIG_IGN); // Don't choke if a client process exit
    (void) signal(SIGHUP, (SIG_PF) CatchPhone);
    DUNGEON dungeon;
    ugly_ugly = &dungeon;	// Bug in g++ 1.36.1
    dungeon.play(&gate);
    return 0;
}

// @@@ void CatchPhone()

static void CatchPhone()
{
    global.SetEmergency();
    bug.warning("::CatchPhone", "Catched a SIGHUP");
}

// @@ emacs
//
// Local Variables:
// mode: c++
// outline-regexp: "// @+ "
// eval: (outline-minor-mode 1)
// End:
