// @ myth - server - play.C
//
// Myth is a multiuser screen oriented adventure game.
// Copyright (C) 1989 Per Abrahamsen
// 
// This file is part of Myth.
// 
// Myth is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 1, or (at your option) any later
// version.
// 
// Myth is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Myth; see the file COPYING.  If not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
// My email address is: amanda@iesd.auc.dk

// @@ include

#include "server/level.h"
#include "common/error.h"
#include "server/watch.h"
#include "common/wait.h"
#include "server/player.h"
#include "server/rnd.h"
#include "server/sleep.h"
#include "common/page.h"
#include "server/hiscore.h"
#include "common/format.h"

// @@ LEVEL

// @@@ bool LEVEL::WatchClock()
//
// Update the watch.
//
// Return TRUE each TickTime second.

bool LEVEL::WatchClock()
{
    if(watch() - LastVisit >= TickTime)
    {
	LastVisit = watch();
	return TRUE;
    }
    else
	return FALSE;
}

// @@@ tick_type LEVEL::MoveAllPlayers(bool* AutoRun)
//
// Parse input to all players on the level, then let them move.
//
// Return:
//    TICK_FIGHT, iff all players have used their fight point.
//    TICK_MOVE, iff a player has used her move point.
//    TICK_TOUCH, iff a player has moved
//    TICK_WAIT otherwise.
//
// Also: Assign AutoRun 

tick_type LEVEL::MoveAllPlayers(bool* AutoRun)
{
    // TRUE, if no players may fight
    bool DoTickFight = TRUE;
    // TRUE, iff any player has used her movement point
    bool DoTickMove = FALSE;
    // TRUE, iff any player has played
    bool DoTickTouch = FALSE;
    
    // Parse input from players
    for(int where = 0; where < party.length(); where++)
    {
	party.reset(where);
	PLAYER* player = party.current();

	if(player->dead())
	    DoTickTouch = *AutoRun = TRUE;
	else if(!player->good())
	{
	    bug.warning("LEVEL::MoveAllPlayers",
			format("bad I/O for %s",
			       print(player, PRINT_MONSTER, player)));
	    player->sleep(this, 0L);
	    DoTickTouch = TRUE;
	}
	else if(player->busy())
	{
	    player->parse();

	    switch(player->answer(this))
	    {
	    case Q_VALUE_PARSING:
		break;
	    case Q_VALUE_DONE:
		DoTickTouch = TRUE;
		player->retry();
		break;
	    case Q_VALUE_AGAIN:
		if(party.length() == 1 || !player->CheckWaiting())
		    DoTickTouch = *AutoRun = TRUE;
		break;
	    case Q_VALUE_ERROR:
	    default:
		bug.error("LEVEL::MoveAllPlayers", "QUESTION::play garbage");
	    }
	    player->ClearInterrupt();
	}
	else
	    bug.error("LEVEL::MoveAllPlayers", "NULL question");
	if(player->MayFight())
	    DoTickFight = FALSE;
	if(!player->MayMove())
	    DoTickMove = TRUE;
    }
    // Return type of tick.
    if(DoTickFight || (DoTickMove && party.length() < 2))
	return TICK_FIGHT;
    else if(DoTickMove)
	return TICK_MOVE;
    else if(DoTickTouch)
	return TICK_TOUCH;
    else
	return TICK_WAIT;
}

// @@@ void LEVEL::MoveAllMonsters()
//
// Reset the movepoint for each monster, and let them spend it (by
// moving around in the dungeon). To be able to use the movepoints,
// the monster must have some fightpoint left.

void LEVEL::MoveAllMonsters()
{
    // Reset the move points
    zoo.reset();
    while(zoo.current() != NULL)
    {
	zoo.current()->TickMove();
	zoo.next();
    }
    // Keep moving the monsters, while they can be moved
    bool again;			// TRUE, iff no monster may move
    do
    {
	again = FALSE;
	
	for(int where = 0; where < zoo.length(); where++)
	{
	    zoo.reset(where);

	    if(zoo.current()->play(this) == CONTINUE)
		again = TRUE;
	}
    }
    while(again);
}

// @@@ void LEVEL::CleanUpPlayerQueue(WAIT* wait,
//
// Remove monsters who have moved away from the level from the list of
// players on the level. Also remove the players who have finished
// playing from the "wait for input" queue.
//
// This *must* be called before CleanUpMonsterQueue.

void LEVEL::CleanUpPlayerQueue(WAIT* wait, SLEEP* sleep, HI_SCORE* HiScore)
{
    // Clean up list of active players
    party.reset();
    while(party.remain() > 0)
    {
	PLAYER* player = party.current();
	
	if(player->done() || player->escape() || player->dead())
	{
	    // Show HiScore
	    PAGE page;
	    HiScore->page(&page);
	    player->page(&page);
	    // Do not wait for input from dead players
	    player->DisConnect(wait);
	    party.remove();
	}
	else if(player->depth() != depth())
	    party.remove();
	else if(player->sleeping())
	{
	    sleep->add(player);
	    PAGE page;
	    sleep->page(&page);
	    player->page(&page);
	    player->DisConnect(wait);
	    sleepers.add(player);
	    party.remove();
	}
	else
	    party.next();
    }
    // Clean up list of sleeping players
    sleepers.reset();
    while(sleepers.remain() > 0)
    {
	PLAYER* player = sleepers.current();

	if(player->depth() != depth() || player->dead())
	{
	    sleep->remove(player->identity());
	    sleepers.remove();
	}
	else
	    sleepers.next();
    }
}

// @@@ void LEVEL::CleanUpMonsterQueue(MONSTER_LIST* export)
//
// Remove monsters who have moved away from the level from the list of
// monsters on the level.

void LEVEL::CleanUpMonsterQueue(MONSTER_LIST* export)
{
    zoo.reset();
    
    while(zoo.current() != NULL)
    {
	MONSTER* monster = zoo.current();
	
	if(monster->depth() != depth() || monster->dead())
	{
	    if(monster->depth() != depth())
		monster->leave(this);
	    
	    export->append(monster);
	    zoo.remove();
	}
	else
	    zoo.next();
    }
}

// @@@ bool LEVEL::play(MONSTER_LIST* export,

bool LEVEL::play(MONSTER_LIST* export, WAIT* wait, SLEEP* sleep,
		 HI_SCORE* HiScore, force flag)
{
    // TRUE, iff some player is automatically moving
    bool AutoRun = FALSE;

    // Read the input from players, and see what happens.
    tick_type TickType = MoveAllPlayers(&AutoRun);

    // Each TickTime second, we want to allow monsters to fight,
    // but not if there only is one player on the level.
    if(WatchClock() && party.length() != 1)
	TickType = TICK_FIGHT;

    // Check if monsters should be moved
    if(TickType != TICK_WAIT || flag == FORCE)
    {
	// Check if monsters should be allowed to fight again
	if(TickType == TICK_FIGHT)
	{
	    // Allow all monsters to fight
	    zoo.reset();
	    while(zoo.current() != NULL)
	    {
		zoo.current()->TickFight();
		zoo.next();
	    }

	    // Create new monsters
	    if(frequence.NewMonster())
	    {
		MONSTER* monster =
		    frequence.CreateMonster(experience());

		AddMonster(monster);
		monster->move(this, locate(frequence.locate()));
		monster->NewLevel(depth(), SYMBOL_ERROR);
	    }
	}

	// Let the monsters move around
	if(TickType != TICK_TOUCH)
	    MoveAllMonsters();

	// Remove players from the level or the game
	CleanUpPlayerQueue(wait, sleep, HiScore);

	// Remove monsters from the level or the game
	CleanUpMonsterQueue(export);
	
	// Make delayed events happen before we update the players display
	event.update(this);

	// Update all players
	for(int where = 0; where < party.length(); where++)
	{
	    party.reset(where);
	    party.current()->update(this);
	}
    }
    // When the last player has left the level, start a count down
    if(party.length() > 0)
	CountDown = ERROR;
    else if(CountDown == ERROR)
	CountDown = rnd(MinLevelLife, MaxLevelLife);

    return AutoRun;
}

// @@ emacs
//
// Local Variables:
// mode: c++
// outline-regexp: "// @+ "
// eval: (outline-minor-mode 1)
// End:
