// @ myth - server - thing.h
//
// Myth is a multiuser screen oriented adventure game.
// Copyright (C) 1989 Per Abrahamsen
// 
// This file is part of Myth.
// 
// Myth is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 1, or (at your option) any later
// version.
// 
// Myth is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Myth; see the file COPYING.  If not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
// My email address is: amanda@iesd.auc.dk

// @@ begin

#ifndef THING_HEADER
#define THING_HEADER

// @@ include

#include "server/object.h"

struct RAY;

// @@ const

static const int MaxValue = 25;

// @@ THING

class THING : public OBJECT
{
    //
    // Data
    //
private:
    char MyLetter;		// What inventory letter is used
    void create(long count = 1L); // Initialize thing
protected:
    long MyCount;		// Number of subitems
    char MyValue;		// How good is it?
    //
    // Identification
    //
public:    
    char letter()		// Letter in inventory
    { return MyLetter; }
    long count()		// #subitems
    { return MyCount; }
    bool IsThing()
    { return TRUE; }
    comp compare(OBJECT* object);
    STRING print(print_flag flag, long count, PLAYER* player);
    virtual const char* PrintValue(); // Good, bad, perfect or useless
    virtual const char* PrintActive(print_flag) // Activated?
    { return NULL; }
    bool CanRemove(MONSTER* monster); // Is it cursed?
    //
    // Value
    //
protected:
    virtual int BaseValue()	// Value of unidentified object
    { return 1; }
    virtual int ClassValue()	// Value of unidentified class members
    { return BaseValue(); }
public:
    long worth(MONSTER* monster);
    //
    // Visibility
    //
private:
    bool invisible;		// Ot even invisible?
public:
    bool visible()
    { return !invisible; }
    void SetVisible()
    { invisible = FALSE; } 
    void ClearVisible()
    { invisible = TRUE; }
    //
    // Holyness
    //
protected:
    holy_type holy;		// Holyness of item?
public:    
    virtual const char* PrintHoly(); // Cursed or blessed?
    bool IsCursed()
    { return holy == CURSED; }
    bool IsBlessed()
    { return holy == BLESSED; }
    bool IsNormal()
    { return holy == NORMAL; }
    void curse()		// Curse the object
    { holy = CURSED; }
    void bless()		// Or bless it
    { holy = BLESSED; }
    void RemoveCurse()
    { if(holy == CURSED) holy = NORMAL; }
    //
    // Action
    //
protected:
    virtual void effect(LEVEL*, MONSTER*); // Special effect when applied
    virtual QUESTION* DoEffect(LEVEL* level, PLAYER* player, CHANNEL*);
    void TakeOff(MONSTER* monster); // Stop wearing it
    virtual void wear(MONSTER* monster); // Start wearing it
public:
    void damage();		// Damage the object
    void ChangeValue(int change)
    { MyValue = min(25, max(1, MyValue + change)); }
    cont event(LEVEL*, coor, MONSTER* = NULL)
    { return MyValue == 0 ? DONE : CONTINUE; }
    void MoveAway(LEVEL* level, coor pos, int length);
    //
    // Administration
    //
public:
    void add(OBJECT* object)	// Add subitems in two objects
    { MyCount += object->count(); }
    void SetLetter(char letter)	// Select letter
    { MyLetter = letter; }
    void save(BINARY* binary);
    THING(BINARY* binary);
    THING(PARSE* parse);
    THING(THING* thing, long count);
    THING(long count = 1L);
};

// @@ MAGIC

class MAGIC : public THING
{
public:
    bool IsMagic()
    { return TRUE; }
    const char* name();
    const char* PluralName();
    virtual const char* MagicName()
    { return "panic"; }
    MAGIC(BINARY* binary) : (binary) { }
    MAGIC(PARSE* parse) : (parse) { }
    MAGIC(MAGIC* thing, long count) : (thing, count) { }
    MAGIC(long count = 1L) : (count)  { }
};    

// @@ TORCH
 
class TORCH : public THING
{
    long last;			// Last turned on
    long left;			// Time left
    void create();
    long time();		// time for next tick
protected:
    int ClassValue()
    { return 5; }
public:
    symbol_type symbol()
    { return SYMBOL_WAND; }
    object_type type()
    { return WAND_TORCH; }
    const char* name()
    { return "torch"; }
    const char* PluralName()
    { return "torches"; }
    const char* PluralClassName()
    { return "torches"; }
    long size()
    { return (10L + 2L * MyValue) * MyCount; }
    long weight()
    { return (10L + 2L * MyValue) * MyCount; }
    comp compare(OBJECT* object);
    bool light()		// Is it burning?
    { return last != 0L; }
    const char* PrintActive(print_flag);
    void apply(LEVEL* level, MONSTER* monster); // Turn On/Off
    cont event(LEVEL* level, coor pos, MONSTER* monster = NULL);
    OBJECT* split(long arg)
    { MyCount -= arg; return new TORCH(this, arg); }
    void save(BINARY* binary);
    TORCH(BINARY* binary);
    TORCH(PARSE* parse) : (parse) { create(); }
    TORCH(TORCH* thing, long count) : (thing, count) { create(); }
    TORCH(long count = 1L) : (count)  { create(); }
};    

// @@ WAND
 
class WAND : public MAGIC
{
private:
    void cleanup(MONSTER* monster); // Put back in pack or destroy
protected:
    virtual RAY* RayCursed();
    virtual RAY* RayNormal();
    virtual RAY* RayBlessed();
    int BaseValue()		// All wands is worth the same
    { return 300; }
public:
    bool IsWand()
    { return TRUE; }
    symbol_type symbol()
    { return SYMBOL_WAND; }
    const char* ClassName()
    { return "wand"; }
    long size()
    { return 75L * MyCount; }
    long weight()
    { return 50L * MyCount; }
    void apply(LEVEL* level, MONSTER* monster); // Zap
    QUESTION* DoApply(LEVEL*, PLAYER*, CHANNEL* channel);
    WAND(BINARY* binary) : (binary) { }
    WAND(PARSE* parse) : (parse) { }
    WAND(WAND* thing, long count) : (thing, count) { }
    WAND(long count = 1L) : (count)  { }
};    

// @@ POTION
 
class POTION : public MAGIC
{
protected:
    int ClassValue()
    { return 25; }
public:
    bool IsPotion()
    { return TRUE; }
    symbol_type symbol()
    { return SYMBOL_POTION; }
    const char* ClassName()
    { return "potion"; }
    long size()
    { return 15 * MyCount; }
    long weight()
    { return 25L * MyCount; }
    void apply(LEVEL* level, MONSTER* monster);	// Quaff
    POTION(BINARY* binary) : (binary) { }
    POTION(PARSE* parse) : (parse) { }
    POTION(POTION* thing, long count) : (thing, count) { }
    POTION(long count = 1L) : (count)  { }
};    

// @@ SCROLL
 
class SCROLL : public MAGIC
{
    void read(LEVEL* level, MONSTER* monster);
protected:
    int ClassValue()
    { return 50; }
public:
    bool IsScroll()
    { return TRUE; }
    symbol_type symbol()
    { return SYMBOL_SCROLL; }
    const char* ClassName()
    { return "scroll"; }
    long size()
    { return 40L * MyCount; }
    long weight()
    { return 5L * MyCount; }
    void apply(LEVEL* level, MONSTER* monster); // Read
    QUESTION* DoApply(LEVEL* level, PLAYER* player, CHANNEL* channel);
    SCROLL(BINARY* binary) : (binary) { }
    SCROLL(PARSE* parse) : (parse) { }
    SCROLL(SCROLL* thing, long count) : (thing, count) { }
    SCROLL(long count = 1L) : (count)  { }
};    

// @@ RING

class RING : public MAGIC
{
protected:
    int ClassValue()
    { return 150; }
public:
    bool IsRing()
    { return TRUE; }
    symbol_type symbol()
    { return SYMBOL_RING; }
    const char* ClassName()
    { return "ring"; }
    long size()
    { return 1L * MyCount; }
    long weight()
    { return 10L * MyCount; }
    void apply(LEVEL* level, MONSTER* monster); // Put on finger
    RING(BINARY* binary) : (binary) { }
    RING(PARSE* parse) : (parse) { }
    RING(RING* thing, long count) : (thing, count) { }
    RING(long count = 1L) : (count)  { }
};

// @@ AMULET

class AMULET : public MAGIC
{
protected:
    int ClassValue()
    { return 5000; }
public:
    symbol_type symbol()
    { return SYMBOL_RING; }
    const char* ClassName()
    { return "amulet"; }
    long size()
    { return 10L * MyCount; }
    long weight()
    { return 20L * MyCount; }
    void apply(LEVEL* level, MONSTER* monster); // Put on neck
    AMULET(BINARY* binary) : (binary) { }
    AMULET(PARSE* parse) : (parse) { }
    AMULET(AMULET* thing, long count) : (thing, count) { }
    AMULET(long count = 1L) : (count)  { }
};

// @@ WEAPON

class WEAPON : public THING
{
protected:
    int ClassValue()
    { return BaseValue(); }
public:
    bool IsWeapon()
    { return TRUE; }
    symbol_type symbol()
    { return SYMBOL_WEAPON; }
    void apply(LEVEL* level, MONSTER* monster); // Wield
    WEAPON(BINARY* binary) : (binary) { }
    WEAPON(PARSE* parse) : (parse) { }
    WEAPON(WEAPON* thing, long count) : (thing, count) { }
    WEAPON(long count = 1L) : (count)  { }
};

// @@ ARMOR

class ARMOR : public THING
{
protected:
    // Change power && type of attack
    virtual int protect(int, damage_type*)
    { return 0; }
    int ClassValue()
    { return BaseValue(); }
public:
    bool IsArmor()
    { return TRUE; }
    symbol_type symbol()
    { return SYMBOL_ARMOR; }
    void apply(LEVEL* level, MONSTER* monster); // Wear
    cont absorb(int* power, damage_type* type,
		OBJECT* object, MONSTER* attack, MONSTER* defence);
    ARMOR(BINARY* binary) : (binary) { }
    ARMOR(PARSE* parse) : (parse) { }
    ARMOR(ARMOR* thing, long count) : (thing, count) { }
    ARMOR(long count = 1L) : (count)  { }
};    

// @@ FOOD
 
class FOOD : public THING
{
    void eat(LEVEL* level, MONSTER* monster);
protected:
    virtual int FoodValue()	// Value when eaten
    { return 0; }
    int BaseValue()
    { return FoodValue() / 30; }
public:
    bool IsFood()
    { return TRUE; }
    symbol_type symbol()
    { return SYMBOL_FOOD; }
    const char* PrintHoly();
    void apply(LEVEL* level, MONSTER* monster);
    QUESTION* DoApply(LEVEL* level, PLAYER* player, CHANNEL* channel);
    FOOD(BINARY* binary) : (binary) { }
    FOOD(PARSE* parse) : (parse) { }
    FOOD(FOOD* thing, long count) : (thing, count) { }
    FOOD(long count = 1L) : (count)  { }
};    

// @@ end

#endif THING_HEADER

// @@ emacs
//
// Local Variables:
// mode: c++
// outline-regexp: "// @+ "
// eval: (outline-minor-mode 1)
// End:
